##############################################
# Name:           lines2tubes
# Author :        Didier Bur
# Description:    Create cylinders with selected edges 
# Usage:          select lines, arcs,circles,curves, select options and click OK
# Date:           2005,13,01
# Revised:        2005,17,01: added followme's where applicable
#                     2016 August: entirely rewritten, profile, mat, soft edges, angle added
##############################################

# Lines2Tubes::Li2Tu.new(Sketchup.active_model.selection)
# f=Sketchup.active_model.selection[0]
# ents=Sketchup.active_model.entities

$l2tDebug=false

module Lines2Tubes
	
	@fProfile=nil
	@tubeCount=1
	@mtubeCount=1
	
	def self.fProfile()
		@fProfile
	end
	
	def self.tubeCount()
		@tubeCount
	end
	def self.tubeCount=(x)
		@tubeCount=x
	end
	def self.mtubeCount()
		@mtubeCount
	end
	def self.mtubeCount=(x)
		@mtubeCount=x
	end
	
	def self.l2t_translate()
			@path = Sketchup.find_support_file("Plugins")+"/DBUR_Lines2Tubes/"
			locale=Sketchup.get_locale.upcase
			langFile=@path+"Resources/Lines2Tubes_"+locale+".lingvo"
			if !File.exist?(langFile)
				langFile=@path+"Resources/Lines2Tubes_EN-US.lingvo"
			end
			#Build translation hash
			aT={}
			texts=IO.readlines(langFile)
			texts.each do |line|
				kv=line.split("<==>")
				aT[kv[0]] = kv[1].chop if kv.length==2
			end
			return aT
		end
		
	# Translate ASAP
	L2T = self.l2t_translate()

	# Return max number of intersecting edges in a path
	def self.typePath(edges)
	a=[]
	edges.each do |e|
		next if e.class!=Sketchup::Edge
		a.push(e.start.edges.length)
		a.push(e.end.edges.length)
	end
	return a.max
	end


	# Check for closed or open path
	def self.openPath?(edgesPath)
		extr=[]
		edgesPath.each { |e| extr.push(e) if e.start.edges.length==1; extr.push(e) if e.end.edges.length==1 } 
		extr.uniq!
	   case extr.length
		when 0
			 return false
		when 1
			 return false
		when 2
			return true
		else
			return false
	    end
	end
	
	
	# Set user face profile for extrusion
	def self.setFaceProfile()
		sel=Sketchup.active_model.selection
		if !sel.empty?
			if sel.length==1 
				if sel[0].class==Sketchup::Face
					@fProfile=sel[0]
					UI.messagebox(Lines2Tubes::L2T["Face profile set."])
				else
					@fProfile=nil
					UI.messagebox(Lines2Tubes::L2T["Incorrect selection."]  + "\n" + Lines2Tubes::L2T["Face profile not set."])
				end
			else
				@fProfile=nil
				UI.messagebox(Lines2Tubes::L2T["Please select only one face."])
			end
		else
			@fProfile=nil
			UI.messagebox(Lines2Tubes::L2T["No selection."] + "\n" + Lines2Tubes::L2T["Please select a face."])
		end
		sel.clear
	end
	
	
	def self.faceCentroid(f)
		xs=0.0
		ys=0.0
		zs=0.0
		f.vertices.each do |v|
			xs += v.position.x
			ys += v.position.y
			zs += v.position.z
		end
		c=Geom::Point3d.new(xs/f.vertices.length,ys/f.vertices.length,zs/f.vertices.length)
		return c
	end
 
 
	 # Copy a face loop, outer or inner
	def self.copyLoop(lop,vec,ents)
		@divCurve=nil
		ol=lop
		newEdges=[]
		edge4Face=nil
		olEdges=ol.edges
		olEdges.each { |e|
			if !e.curve
				#copy edge
				newEdges.push([e.start.position.offset(vec),e.end.position.offset(vec)])
				else
				# copy curve
				newCurve=[]
				e.curve.edges.each { |ce|
					newCurve.push(ce.start.position.offset(vec))
					newCurve.push(ce.end.position.offset(vec))
					newCurve.uniq!
				}
				newEdges.push(newCurve) if !newEdges.include?(newCurve)
			end
		}
		# Copy loop
		newEdges.each { |e|
			if e.length==2
				# regular edge
				edge4Face=ents.add_line(e[0],e[1])
			else
				# curve
				curv=ents.add_curve(e)
				edge4Face=curv[0]
				# store curve for divide
				@divCurve=curv if !openPath?(curv)
			end
		}
		# Create face
		nFace=edge4Face.find_faces()
		if nFace==1
			return ents[ents.length-1]
		else
			# Divide inner face by creating an edge in it
			divideFace(@divCurve,ents)
			return nil
		end
	end

	
	# Create edge and erase it, so 2 inner faces are created then erased
	def self.divideFace(divCurve,ents)
		newEdge=ents.add_line(divCurve[0].start.position,divCurve[1].end.position)
		newEdge.erase! if newEdge
	end
	
	
	# Copy face from its fCentroid to ORIGIN (vec), to context ents
	def self.copyFace(f,vec,ents)
		outer_face=copyLoop(f.outer_loop,vec,ents)
		puts("outer face: " + outer_face.inspect) if $l2tDebug
		
		inner_faces = []
		if f.loops.length > 1
			il = f.loops
			il.shift
			puts("il: " + il.inspect) if $l2tDebug
			il.each do |loop|
				inner_face=copyLoop(loop,vec,ents)
				puts("inner face: " + inner_face.inspect) if $l2tDebug
				inner_faces.push(inner_face) if inner_face
			end
			puts("outer face: " + outer_face.inspect) if $l2tDebug
			puts inner_faces.inspect if $l2tDebug
			inner_faces.each { |f| f.erase! }
		end
		return outer_face
    end
    
    
	# Orient face so its normal is pointing Z down, assuming centroid is at ORIGIN
	def self.orientFaceToZ(f)
		fc=self.faceCentroid (f)
		tSource=Geom::Transformation.new(fc, f.normal).inverse
		tTarget=Geom::Transformation.new(ORIGIN, Z_AXIS.reverse)
		Sketchup.active_model.entities.transform_entities(tSource,f)
		Sketchup.active_model.entities.transform_entities(tTarget,f)
	end
	
	
	# Copy faceProfile to ents entities, and orient it at starting point of edge e, return new face
	def self.orientFaceToEdge(ents,e)
		f=self.fProfile
		puts("Face profile: " + f.inspect) if $l2tDebug
		return if f==nil
		fCentroid=faceCentroid (f)
		vecCentroidOrigin=fCentroid.vector_to(ORIGIN)
		# Copy face from its centroid to ORIGIN
		newFace=copyFace(f,vecCentroidOrigin,ents)
		# Make it pointing Z down
		self.orientFaceToZ(newFace)
		# Now fCentroid is at ORIGIN
		vecCentroidEdge=ORIGIN.vector_to(e.start.position)
		tTranslate=Geom::Transformation.translation(vecCentroidEdge)
		fNorm=newFace.normal
		tEdge=Geom::Transformation.new(ORIGIN, e.line[1])
		ents.transform_entities(tEdge,newFace)
		ents.transform_entities(tTranslate,newFace)
		return newFace
	end
	
class Li2Tu

	attr_accessor :model, :sel, :ss_ignored, :ss_edges, :diam, :ray, :segs,  :initAngle, :opt, :mat, :faceProfile, :smoothEdges, :mainGrp, :mGroup, :tubeGrp, :tGroup, :mEnts, :tEnts, :mainEnts, :matsH, :trunk, :nTubes

	@@diam=10.cm
	@@segs=24
	@@initAngle=0.degrees
	@@profile=Lines2Tubes::L2T["Polygon"]
	@@opt=Lines2Tubes::L2T["Yes"]
	@@mainGrp=Lines2Tubes::L2T["Yes"]
	@@tubeGrp=Lines2Tubes::L2T["No"]
	@@smoothEdges=Lines2Tubes::L2T["Yes"]
	@@mat=Lines2Tubes::L2T["Default"]
	@@faceProfile=nil
	
	def initialize(sel)
		@model =Sketchup.active_model
		@mEnts = @model.active_entities
		@sel = sel
		@nTubes=0
		@mat=nil
		@trunk=[]
		Sketchup.set_status_text("")
		if @model.selection.empty?
			UI.messagebox(Lines2Tubes::L2T["No selection."])
			return
		end
	end
	
	def opt()
		@@opt
	end
	
	def faceProfile()
		@@faceProfile
	end
	
	def smoothEdges
		@@smoothEdges
	end
	
	
	def optionsDialog()
		# Materials list
		list2=[]
		list2=@model.materials.collect { |m| m.display_name }
		list2.push(Lines2Tubes::L2T["Default"])
		list2.push(Lines2Tubes::L2T["From profile"])
		if @@mat==Lines2Tubes::L2T["Default"]
			@mat=nil
		elsif @@mat==Lines2Tubes::L2T["From profile"]
			@mat=Lines2Tubes.fProfile.material if Lines2Tubes.fProfile
			@mat=nil if !Lines2Tubes.fProfile
		else
			@mat=@matsH[@@mat]
		end
		# Dropdown lists
		drops=[[Lines2Tubes::L2T["Polygon"],Lines2Tubes::L2T["Face"]].join("|"),[Lines2Tubes::L2T["Yes"],Lines2Tubes::L2T["No"]].join("|"),[Lines2Tubes::L2T["Yes"],Lines2Tubes::L2T["No"]].join("|"),[Lines2Tubes::L2T["Yes"],Lines2Tubes::L2T["No"]].join("|"),[Lines2Tubes::L2T["Yes"],Lines2Tubes::L2T["No"]].join("|"),list2.join("|")]
		prompts = [Lines2Tubes::L2T["Profile"]+": ",Lines2Tubes::L2T["Follow me if applicable"]+": ",Lines2Tubes::L2T["Group all tubes"]+": ",Lines2Tubes::L2T["Each tube is a group"] +": ",Lines2Tubes::L2T["Smooth edges"] +": ",Lines2Tubes::L2T["Material"]+": ",Lines2Tubes::L2T["Diameter"]+": ", Lines2Tubes::L2T["Segments"]+":  ", Lines2Tubes::L2T["Profile angle"]+":  "]
		values = [@@profile,@@opt,@@mainGrp,@@tubeGrp,@@smoothEdges, @@mat,@@diam,@@segs,@@initAngle.radians]
		# Dialog
		results = inputbox prompts, values,drops, Lines2Tubes::L2T["Options"]
		return if not results
		@@profile=results[0]
		@@opt = results[1]
		@@mainGrp = results[2]
		@@tubeGrp = results[3]
		@@smoothEdges = results[4]
		@@mat = results[5]
		@@diam = results[6]
		@@segs = results[7]
		@@initAngle=results[8].degrees
		# Set some variables, groups and entities
		@@segs=3 if @@segs<3
		@ray=@@diam/2.0
		@ss_edges=[]
		@ss_ignored=[]
		if @@mat==Lines2Tubes::L2T["Default"]
			@mat=nil
			puts("mat: " + @mat.inspect) if $l2tDebug
		elsif @@mat==Lines2Tubes::L2T["From profile"]
			@mat=Lines2Tubes.fProfile.material if Lines2Tubes.fProfile
			@mat=nil if !Lines2Tubes.fProfile
		else
			@mat=@matsH[@@mat]
		end
		if @@profile==Lines2Tubes::L2T["Face"] and !Lines2Tubes.fProfile
			UI.messagebox(Lines2Tubes::L2T["Face profile not set."])
			Sketchup.active_model.commit_operation
			return
		end
		# Create main group if requested
		if @@mainGrp==Lines2Tubes::L2T["Yes"]
			@mGrp=@mEnts.add_group
			@mainEnts=@mGrp.entities
			@mGrp.name=Lines2Tubes::L2T["Main tube group"] + "  #"+Lines2Tubes.mtubeCount.to_s
			Lines2Tubes.mtubeCount += 1
		else
			@mainEnts=@mEnts
		end
	end
	
	
	# Check if a set of connected edges has 2 free end and no YTX junctions
	def freePath?(edgesArr)
		# TYX junctions must be skipped
		return false if (joiningEdges(edgesArr) > 2)
		# Find 2 free ends
		free_Ends=freeEnds(edgesArr)
		if free_Ends.length==2
			puts("Free path: true" ) if $l2tDebug
			return true
		else
			puts("Free path: false" ) if $l2tDebug
			return false
		end
	end
	
	# Return free end edges of an array of edges
	def freeEnds(edgesArr)
		free_Ends=[]
		edgesArr.each { |e|
			next if e.class==Sketchup::Face
			free_Ends.push(e) if (freeVertex?(e.start) or freeVertex?(e.end))
		}
		return free_Ends
	end
	
	# Check if a vertex is free, return bool
	def freeVertex?(v)
		if v.edges.length==1
			return true
		else
			return false
		end
	end
	
	
	# Return free vertex of edge e, or nil
	def freeVertexEdge(e)
		return e.start if freeVertex?(e.start)
		return e.end if freeVertex?(e.end)
		return nil
	end
	
	def edgeOfFace?(edgesArr)
		result =false
		edgesArr.delete_if { |e| e.class!=Sketchup::Edge }
		edgesArr.each { |e| result=true if e.faces.length>0 }
		return result
	end
	
	# Check if edgesArr makes a free closed loop (excluding Face connection)
	def freeClosedLoop?(edgesArr)
		result=true
		edgesArr.each { |e| result=false if e.class==Sketchup::Face }
		if result
			result=false if !closedLoop?(edgesArr)
		end
		puts("Free closed loop: " + result.inspect) if $l2tDebug
		return result
	end
	
	# Check if edgesArr makes a closed loop (including Face connection)
	def closedLoop?(edgesArr)
		# No free ends
		fe=freeEnds(edgesArr)
		# Max junctions=2
		nj=joiningEdges(edgesArr)
		if fe.length==0 and nj==2
			puts("Closed loop true") if $l2tDebug
			return true
		else
			puts("Closed loop false") if $l2tDebug
			return false
		end
	end
	
	# Check if edgesArr makes a single face loop, return false otherwise
	def singleFaceLoop?(edgesArr)
		result=false
		f=[]
		edgesArr.each { |e| f.push(e) if e.class==Sketchup::Face }
		return f if (f.length==1)
		return result
	end
	
	# Check if edgesArr makes multiple loops of more than 1 face; return false or faces array
	def multipleFacesLoop?(edgesArr)
		result=false
		f=[]
		edgesArr.each { |e| f.push(e) if e.class==Sketchup::Face }
		return f if f.length>1
		return result
	end
	
	# return an array of all loops of faces array
	def facesLoops(facesArr)
		loups=[]
		facesArr.each do |f|
			f.loops.each { |l| loups.push(l.edges) }
		end
		return loups
	end
	
	# Return max number of joining edges in a path
	def joiningEdges(edgesArr)
	a=[]
	edgesArr.each { |e|
		next if e.class==Sketchup::Face
		a.push(e.start.edges.length)
		a.push(e.end.edges.length)
	}
	puts("Junctions max: " + a.max.inspect) if $l2tDebug
	return a.max
	end
	
	# Ignore non-edge and remove from selection
	def parse_edges()
		Sketchup.set_status_text(Lines2Tubes::L2T["Parsing selection..."])
		@sel.each do |x|
			if x.class==Sketchup::Edge
				@ss_edges.push(x)
			else
				@ss_ignored.push(x)
			end
		end
		@ss_edges.uniq!
		puts(@ss_edges.length.inspect + " edges, " + @ss_ignored.length.inspect + " ignored") if $l2tDebug
		@sel.remove(@ss_ignored)
	end
	
	def dispatchEdges()
		@ss_processed = []
		@ss_alone = []
		@ss_curves = []
		@ss_connected = []
		@ss_single_paths = []
		@ss_multiple_paths = []
		@ss_loops = []
		
		@ss_edges.each do |e|
			# Lonely edges
			if e.all_connected.length==1 and not e.curve
				@ss_alone.push(e)
				updateProcessed([e])
			end
			# Free curves, closed or open
			if c=e.curve
				extr=freeEnds(c.edges)
				# Curve of 1 edge
				if (extr.length == 1 and !edgeOfFace?(c.edges) and (e.all_connected.length==c.edges.length))
					@ss_curves.push(c)
					updateProcessed(e.curve.edges)
				end
				if ((extr.length == 0 or  extr.length == 2) and !edgeOfFace?(c.edges) and (e.all_connected.length==c.edges.length))
					#@ss_curves.push(e.curve.edges)
					@ss_curves.push(c)
					updateProcessed(e.curve.edges)
				end
			end
			# Free connected edges sets, made of edges and/or curves
			if e.all_connected.length>1 and !@ss_processed.include?(e) #and !edgeOfFace?(e.all_connected)
				@ss_connected.push(e.all_connected)
				updateProcessed(e.all_connected)
			end
			
		end
		@ss_curves.uniq!
		@ss_connected.uniq!
		@ss_alone.uniq!
		@ss_processed.uniq!
		puts("Processed first run: " + @ss_processed.length.inspect) if $l2tDebug
		
		# Dispatch connected edges sets -> free loops, face loops, single/multiple paths
		@ss_connected.each do |edgesArr|
			allEdges=edgesArr.clone
			allEdges=pathPurgeFaces(allEdges)
			usedEdges=[]
			#is it a free single path?
			if freePath?(edgesArr)
				@ss_single_paths.push(edgesArr)
			#is it a free single loop?
			elsif freeClosedLoop?(edgesArr)
				@ss_loops.push(edgesArr)
				puts("free closed loop found ") if $l2tDebug
				updateProcessed(edgesArr)
			#is it a face single loop?
			elsif (facesArr=singleFaceLoop?(edgesArr))
				puts("single face loop found ") if $l2tDebug
				loups=facesLoops(facesArr)
				loups.each { |l| @ss_loops.push(l) ; l.each { |e| usedEdges.push(e)} }
				loups.each { |l| updateProcessed(l) }
				# Are there remaining edges in edgesArr ?
				if usedEdges.length < allEdges.length
					puts"REMAINING edges" + (usedEdges.length - allEdges.length).inspect if $l2tDebug
					remainingEdges=allEdges-usedEdges
					# pushpull remaining edges
					pushpullEachEdge(remainingEdges,false)
				end
				
			#is it a face multiple loop?
			elsif (facesArr=multipleFacesLoop?(edgesArr))
				puts("multiple faces loops found ") if $l2tDebug
				loups=facesLoops(facesArr)
				loups.each { |l| @ss_loops.push(l) }
				loups.each { |l| updateProcessed(l) }
			
			else
				@ss_multiple_paths.push(edgesArr)
				updateProcessed(edgesArr)
			end
		end
		puts("Connected sets: " + @ss_connected.length.inspect) if $l2tDebug
		puts("Free curves: " + @ss_curves.length.inspect) if $l2tDebug
		puts("Single free paths: " + @ss_single_paths.length.inspect) if $l2tDebug
		puts("Multiple free paths: " + @ss_multiple_paths.length.inspect) if $l2tDebug
		puts("Total loops found: " + @ss_loops.length.inspect) if $l2tDebug
	end
	
	def matsHash()
		@matsH={}
		@model.materials.each do |mat| @matsH[mat.display_name]=mat end
	end
	
	def rotateProfile(prof,e,ents)
		trans = Geom::Transformation.rotation(e.line[0], e.line[1], @@initAngle)
		ents.transform_entities(trans, prof)
	end
	
	def hardenEdges(ents)
		ents.each do |e|
			next if e.class!=Sketchup::Edge
			e.soft=false
			e.smooth=false
		end
	end
	
	# Add edgesArr to @ss_processed (sorted edges array)
	def updateProcessed(edgesArr)
		edgesArr.each { |e| @ss_processed.push(e) if (e.class==Sketchup::Edge and !@ss_processed.include?(e)) }
		puts("Processed edges so far: " + @ss_processed.length.inspect) if $l2tDebug
	end
	
	# Purge faces objects from connected edges array
	def pathPurgeFaces(connectedEdges)
		return connectedEdges.delete_if { |e| e.class!=Sketchup::Edge }
		end
		
	# No follow me, no curves, pushpull each edge individually
	def pushpullEachEdge(pp_edges,selected)
		puts("Pushpulling each edge") if $l2tDebug
		pp_edges.each do |e|
			# skip edges that were not selected
			next if !@sel.contains?(e) and selected
			# else proceed
			vec = e.line[1]
			tube_group=@mainEnts.add_group
			tube_ents=tube_group.entities
			tube_group.name=Lines2Tubes::L2T["tube"] + " #" + Lines2Tubes.tubeCount.to_s if @@tubeGrp==Lines2Tubes::L2T["Yes"]
			
			# Use faceProfile for pushpull/followme
			if @@profile==Lines2Tubes::L2T["Face"]
				base=Lines2Tubes.orientFaceToEdge(tube_ents,e)
				# Rotate profile
				rotateProfile(base,e,tube_ents) if @@initAngle!=0
				# Apply material
				base.material=@mat
				base.followme(e)
				#Smooth/Soft
				puts("tube_ents: " + tube_ents.length.inspect) if $l2tDebug
				hardenEdges(tube_ents) if @@smoothEdges==Lines2Tubes::L2T["No"]
				# Store tube group in a trunk
				@trunk.push(tube_group)
				@nTubes += 1
				Lines2Tubes.tubeCount += 1
			else
				# Use regular polygon for pushpull/followme
				circle = tube_ents.add_circle(e.start.position, vec, @ray, @@segs)
				if circle
					base = tube_ents.add_face(circle)
					if base
						# Rotate profile
						rotateProfile(base,e,tube_ents) if @@initAngle!=0
						# Apply material
						base.material=@mat
						base.reverse! if base.normal.samedirection?(vec)
						#base.pushpull e.length
						base.followme(e)
						@nTubes += 1
						Lines2Tubes.tubeCount += 1
					end
					#Smooth/Soft
					hardenEdges(tube_ents) if @@smoothEdges==Lines2Tubes::L2T["No"]
					# Store tube group in a trunk
					@trunk.push(tube_group)
				end
			end
		end
		puts("Global tubes count: " + Lines2Tubes.tubeCount.inspect) if $l2tDebug
		# Explode all tube groups if requested
		if @@tubeGrp==Lines2Tubes::L2T["No"]
			@trunk.each { |tube| tube.explode if tube.valid? }
		end
	end
	
	# follow me on a set of edges as if it is a single path
	def followMeEdges(pathEdges)
		puts "Processing set of edges" if $l2tDebug		
		############################ Trouver les free ends et le free vertex avant de démarrer le follow me
		free_Ends=freeEnds(pathEdges)
		edge1 = free_Ends[0]
		vec = edge1.line[1]
		
		if freeVertex?(edge1.start)
			ext1=edge1.start.position
			puts "start at start, no reverse" if $l2tDebug
			reverseNeeded=false
		else
			ext1=edge1.end.position
			puts "start at end, reverse" if $l2tDebug
			reverseNeeded=true
		end
		
		all_edges = pathEdges
		tube_group=@mainEnts.add_group
		tube_ents=tube_group.entities
		tube_group.name=Lines2Tubes::L2T["tube"] + " #" + Lines2Tubes.tubeCount.to_s if @@tubeGrp==Lines2Tubes::L2T["Yes"]
		
		# Use faceProfile for pushpull/followme
		if @@profile==Lines2Tubes::L2T["Face"]
			base=Lines2Tubes.orientFaceToEdge(tube_ents,edge1)
			#base.reverse! if !base.normal.samedirection?(vec)
			#base.reverse! if !base.normal.samedirection?(vec) and reverseNeeded
			base.reverse! if reverseNeeded
			# Rotate profile
			rotateProfile(base,edge1,tube_ents) if @@initAngle!=0
			# Apply material
			base.material=@mat
			base.followme(all_edges)
			#Smooth/Soft
			puts("tube_ents: " + tube_ents.length.inspect) if $l2tDebug
			hardenEdges(tube_ents) if @@smoothEdges==Lines2Tubes::L2T["No"]
			# Store tube group in a trunk
			@trunk.push(tube_group)
			# Remove from selection
			@sel.remove(all_edges)
			@nTubes += 1
			Lines2Tubes.tubeCount += 1
		else
			circle = tube_ents.add_circle(ext1, vec, @ray, @@segs)
			if circle
				base = tube_ents.add_face(circle)
				# Apply material
				base.material=@mat
				# Rotate profile
				rotateProfile(base,c.first_edge,tube_ents) if @@initAngle!=0
				if base
					#base.reverse! if !reverseNeeded
					 base.reverse! if !base.normal.samedirection?(vec) and reverseNeeded
					 puts("follow me on free single path") if $l2tDebug
					 base.followme all_edges
					 @nTubes += 1
					 Lines2Tubes.tubeCount += 1
				end
			#Smooth/Soft
			hardenEdges(tube_ents) if @@smoothEdges==Lines2Tubes::L2T["No"]
			# Store tube group in a trunk
			@trunk.push(tube_group)
			end
			# Remove from selection
			@sel.remove(all_edges)
		end
	end
	
	# Follow me requested:
	def lines2tubes()
		
		# Dispatch edges
		Sketchup.set_status_text(Lines2Tubes::L2T["Sorting edges..."])
		dispatchEdges()
		
		Sketchup.set_status_text(Lines2Tubes::L2T["Generating tubes..."])

		# Pushpull lonely edges
		@ss_alone.each do |e|
			vec = e.line[1]
			tube_group=@mainEnts.add_group
			tube_ents=tube_group.entities
			tube_group.name=Lines2Tubes::L2T["tube"] + " #" + Lines2Tubes.tubeCount.to_s if @@tubeGrp==Lines2Tubes::L2T["Yes"]
			
			# Use faceProfile for pushpull/followme
			if @@profile==Lines2Tubes::L2T["Face"]
				base=Lines2Tubes.orientFaceToEdge(tube_ents,e)
				# Rotate profile
				rotateProfile(base,e,tube_ents) if @@initAngle!=0
				# Apply material
				base.material=@mat
				base.followme(e)
				#Smooth/Soft
				puts("tube_ents: " + tube_ents.length.inspect) if $l2tDebug
				hardenEdges(tube_ents) if @@smoothEdges==Lines2Tubes::L2T["No"]
				# Store tube group in a trunk
				@trunk.push(tube_group)
				# Remove from selection
				@sel.remove(e)
				@nTubes += 1
				Lines2Tubes.tubeCount += 1
			else
				circle = tube_ents.add_circle(e.vertices[0], vec, @ray, @@segs)
				if circle
					base = tube_ents.add_face(circle)
					# Apply material
					base.material=@mat
					# Rotate profile
					rotateProfile(base,e,tube_ents) if @@initAngle!=0
					if base
						base.reverse! if base.normal.samedirection?(vec)
						base.followme(e)
					end
				#Smooth/Soft
				hardenEdges(tube_ents) if @@smoothEdges==Lines2Tubes::L2T["No"]
				# Store tube group in a trunk
				@trunk.push(tube_group)
				# Remove from selection
				@sel.remove(e)
				@nTubes += 1
				Lines2Tubes.tubeCount += 1
			end
		end
	end

		# Followme's on curves
		@ss_curves.each do |c|
			edge1 = c.first_edge
			vec = edge1.line[1]
			ext1 = edge1.start.position
			all_edges = c.edges
			tube_group=@mainEnts.add_group
			tube_ents=tube_group.entities
			tube_group.name=Lines2Tubes::L2T["tube"] + " #" + Lines2Tubes.tubeCount.to_s if @@tubeGrp==Lines2Tubes::L2T["Yes"]
			
			# Use faceProfile for pushpull/followme
			if @@profile==Lines2Tubes::L2T["Face"]
				base=Lines2Tubes.orientFaceToEdge(tube_ents,edge1)
				# Rotate profile
				rotateProfile(base,edge1,tube_ents) if @@initAngle!=0
				# Apply material
				base.material=@mat
				base.followme(all_edges)
				#Smooth/Soft
				puts("tube_ents: " + tube_ents.length.inspect) if $l2tDebug
				hardenEdges(tube_ents) if @@smoothEdges==Lines2Tubes::L2T["No"]
				# Store tube group in a trunk
				@trunk.push(tube_group)
				# Remove from selection
				@sel.remove(all_edges)
				@nTubes += 1
				Lines2Tubes.tubeCount += 1
			else
				circle = tube_ents.add_circle(ext1, vec, @ray, @@segs)
				if circle
					base = tube_ents.add_face(circle)
					# Apply material
					base.material=@mat
					# Rotate profile
					rotateProfile(base,c.first_edge,tube_ents) if @@initAngle!=0
					if base
						 base.reverse! if base.normal.samedirection?(vec)
						 puts("follow me on curve") if $l2tDebug
						 base.followme all_edges
						 @nTubes += 1
						 Lines2Tubes.tubeCount += 1
					end
				#Smooth/Soft
				hardenEdges(tube_ents) if @@smoothEdges==Lines2Tubes::L2T["No"]
				# Store tube group in a trunk
				@trunk.push(tube_group)
				end
				# Remove from selection
				@sel.remove(all_edges)
			end
		end
		
		puts("All loops count: " + @ss_loops.length.inspect) if $l2tDebug
		# Followme's on free loops and face loops
		@ss_loops.each do |conn_e|
			puts("Loop length: " + conn_e.length.inspect) if $l2tDebug
			 ext_edges = freeEnds(conn_e)
			first_ed=conn_e[0]
			vec = first_ed.line[1]
			ext1=first_ed.start.position
			
			if freeVertex?(first_ed.start)
				ext1=first_ed.start.position
				puts "start at start, no reverse" if $l2tDebug
				reverseNeeded=false
			else
				ext1=first_ed.end.position
				puts "start at end, reverse" if $l2tDebug
				reverseNeeded=true
			end
					
			tube_group=@mainEnts.add_group
			tube_ents=tube_group.entities
			tube_group.name=Lines2Tubes::L2T["tube"] + " #" + Lines2Tubes.tubeCount.to_s if @@tubeGrp==Lines2Tubes::L2T["Yes"]
					
			# Use faceProfile for pushpull/followme
			if @@profile==Lines2Tubes::L2T["Face"]
				base=Lines2Tubes.orientFaceToEdge(tube_ents,first_ed)
				base.reverse! if !reverseNeeded
				# Rotate profile
				rotateProfile(base,first_ed,tube_ents) if @@initAngle!=0
				# Apply material
				base.material=@mat
				base.followme(conn_e)
				#Smooth/Soft
				puts("tube_ents: " + tube_ents.length.inspect) if $l2tDebug
				hardenEdges(tube_ents) if @@smoothEdges==Lines2Tubes::L2T["No"]
				# Store tube group in a trunk
				@trunk.push(tube_group)
				# Remove from selection
				@sel.remove(conn_e)
				@nTubes += 1
				Lines2Tubes.tubeCount += 1
			else
				# Regular polygon
				circle = tube_ents.add_circle(ext1, vec, @ray, @@segs)
				if circle
					base = tube_ents.add_face(circle)
					base.reverse!  if !(vec.samedirection?(base.normal) and reverseNeeded)
					# Apply material
					base.material=@mat
					# Rotate profile
					rotateProfile(base,first_ed,tube_ents) if @@initAngle!=0
					if base 
						base.followme conn_e
						 @nTubes += 1
						 Lines2Tubes.tubeCount += 1
					end
					#Smooth/Soft
					hardenEdges(tube_ents) if @@smoothEdges==Lines2Tubes::L2T["No"]
				# Remove from selection
				@sel.remove(conn_e)
				# Store tube group in a trunk
				@trunk.push(tube_group)
			end
		end
	end


		# Follow me's on free single paths
		@ss_single_paths.each do |singleFreePath|
			puts "Processing free single path" if $l2tDebug
			ext_edges = freeEnds(singleFreePath)
			edge1 = ext_edges[0]
			vec = edge1.line[1]
			
			if freeVertex?(edge1.start)
				ext1=edge1.start.position
				puts "start at start, no reverse" if $l2tDebug
				reverseNeeded=false
			else
				ext1=edge1.end.position
				puts "start at end, reverse" if $l2tDebug
				reverseNeeded=true
			end
			
			all_edges = singleFreePath
			tube_group=@mainEnts.add_group
			tube_ents=tube_group.entities
			tube_group.name=Lines2Tubes::L2T["tube"] + " #" + Lines2Tubes.tubeCount.to_s if @@tubeGrp==Lines2Tubes::L2T["Yes"]
			
			# Use faceProfile for pushpull/followme
			if @@profile==Lines2Tubes::L2T["Face"]
				base=Lines2Tubes.orientFaceToEdge(tube_ents,edge1)
				base.reverse! if reverseNeeded
				# Rotate profile
				rotateProfile(base,edge1,tube_ents) if @@initAngle!=0
				# Apply material
				base.material=@mat
				base.followme(all_edges)
				#Smooth/Soft
				puts("tube_ents: " + tube_ents.length.inspect) if $l2tDebug
				hardenEdges(tube_ents) if @@smoothEdges==Lines2Tubes::L2T["No"]
				# Store tube group in a trunk
				@trunk.push(tube_group)
				# Remove from selection
				@sel.remove(all_edges)
				@nTubes += 1
				Lines2Tubes.tubeCount += 1
			else
				circle = tube_ents.add_circle(ext1, vec, @ray, @@segs)
				if circle
					base = tube_ents.add_face(circle)
					# Apply material
					base.material=@mat
					# Rotate profile
					rotateProfile(base,c.first_edge,tube_ents) if @@initAngle!=0
					if base
						base.reverse! if !reverseNeeded
						 #base.reverse! if base.normal.samedirection?(vec)
						 puts("follow me on free single path") if $l2tDebug
						 base.followme all_edges
						 @nTubes += 1
						 Lines2Tubes.tubeCount += 1
					end
				#Smooth/Soft
				hardenEdges(tube_ents) if @@smoothEdges==Lines2Tubes::L2T["No"]
				# Store tube group in a trunk
				@trunk.push(tube_group)
				end
				# Remove from selection
				@sel.remove(all_edges)
			end
		end
		
		# Follow me's on free multiple paths
		@ss_multiple_paths.each do |multipleFreePath|
			puts "Processing multiple path" if $l2tDebug
			mPath=multipleFreePath.clone
			free_Ends=freeEnds(mPath)
			puts("Free ends: " + free_Ends.length.inspect) if $l2tDebug
			nRemain=mPath.length
			if free_Ends.length>0
				# search for a path
					stop=false
				while !stop and free_Ends.length > 0
					firstEdge=free_Ends[0]
					free_Ends.delete(firstEdge)
					puts("Free ends: " + free_Ends.length.inspect) if $l2tDebug
					pathEdges=findSinglePathInMultiplePath(firstEdge,mPath)
					pathEdges.each { |e| free_Ends.delete(e ) if free_Ends.include?(e) }
					mPath-=pathEdges
					nRemain-=pathEdges.length
					nRemain-=1
					puts("edgesUsed: " + pathEdges.inspect) if $l2tDebug
					puts("nRemain: " + nRemain.inspect) if $l2tDebug
					followMeEdges(pathEdges)
					stop=true if nRemain <=0
				end
				# Pushpull remaing edges
				puts("pushpullEachEdge remaining: " +mPath.length.inspect) if $l2tDebug
				puts("remaining: " +mPath.inspect) if $l2tDebug
				pushpullEachEdge(mPath,false)
			else
				# pushpull each edge
				pushpullEachEdge(mpath,false)
			end
		end
		
		# End of process
		Sketchup.set_status_text("")
		
		# Explode all tube groups if requested
		if @@tubeGrp==Lines2Tubes::L2T["No"]
			puts("Exploding " + @trunk.length.inspect + " individual tubes") if $l2tDebug
			@trunk.each { |tube| tube.explode if tube.valid? }
		end
	end #of def
		
	
	# Tries to extract a path in a multiple path
	def findSinglePathInMultiplePath(firstEdge,edgesArr)
		pathEdges=edgesArr.clone
		edgesUsed=[]
		edgesUsed.push(firstEdge)
		verticesUsed=[]
		stop=false
		firstVertex=freeVertexEdge(firstEdge)
		secondVertex=firstEdge.other_vertex(firstVertex)
		verticesUsed.push(firstVertex)
		verticesUsed.push(secondVertex)
		
		puts("firstEdge 1: " + firstEdge.inspect) if $l2tDebug
		puts("pathEdges 1: " + pathEdges.inspect) if $l2tDebug
		pathEdges.delete(firstEdge)
		puts("pathEdges decremented 1: " + pathEdges.inspect) if $l2tDebug
		
		while !stop
			#find one of the edges sharing second vertex in pathEdges
			nextEdgeData=findRandomEdge(secondVertex, pathEdges) # -> nextEdge, firstV, secondV
			if nextEdgeData
				nextEdge=nextEdgeData[0]
				puts("nextEdge: " + nextEdge.inspect) if $l2tDebug
				firstVertex=nextEdgeData[1]
				secondVertex=nextEdgeData[2]
				# decrement pathEdges and increment vertex
				edgesUsed.push(nextEdge) if !edgesUsed.include?(nextEdge)
				pathEdges.delete(nextEdge)
				puts("pathEdges decremented: " + pathEdges.inspect) if $l2tDebug
			else
				stop=true
			end
			# Stop if we reach another second free end vertex or if secondVertex is yet used
			stop=true if (freeVertex?(secondVertex) or verticesUsed.include?(secondVertex) or pathEdges.length==0)
			verticesUsed.push(secondVertex)
		end
		
		edgesUsed.uniq!
		
		return edgesUsed
	end
	
	
	#find a random edge sharing  vertex v in multiple path pathEdges
	def findRandomEdge(v, pathEdges)
		nextEdge=nil
		pathEdges.each { |e|
			next if e.class==Sketchup::Face
			nextEdge=e if (e.start==v or e.end==v)
		}
		if nextEdge
			if nextEdge.start==v
				firstV=nextEdge.start
				secondV=nextEdge.end
			else
				firstV=nextEdge.end
				secondV=nextEdge.start
			end
			return [nextEdge, firstV, secondV]
		else
			return nil
		end
	end
	
	
end # class Li2Tu

def self.cmdlines2tubes()
	Sketchup.active_model.start_operation(Lines2Tubes::L2T["Lines -> Tubes"],true)
	# Create object
	obj=Lines2Tubes::Li2Tu.new(Sketchup.active_model.selection)
	# Materials H
	obj.matsHash()
	# Dialog
	status=obj.optionsDialog() if !Sketchup.active_model.selection.empty?
	 if status
		# Parse selection -> @ss_edges
		obj.parse_edges()
		# if no follow me => each edge is a tube
		if obj.opt==Lines2Tubes::L2T["No"]
			obj.pushpullEachEdge(obj.ss_edges,true)
		else
			obj.lines2tubes()
		end
		#Message
		Sketchup.set_status_text(obj.nTubes.to_s + " "  + Lines2Tubes::L2T["tubes created,"] + "\n" + obj.ss_ignored.length.to_s + " "  + Lines2Tubes::L2T["objects were invalid and ignored."])
		UI.messagebox(obj.nTubes.to_s + " "  + Lines2Tubes::L2T["tubes created,"] + "\n" + obj.ss_ignored.length.to_s + " "  + Lines2Tubes::L2T["objects were invalid and ignored."])
	else
		Sketchup.set_status_text("Command cancelled.")
	end
	Sketchup.active_model.commit_operation
	puts("Selection: " + obj.sel.length.inspect) if $l2tDebug
end


if( not file_loaded?("lines2tubes.rb") )
    l2t_menu=UI.menu("Tools").add_submenu(Lines2Tubes::L2T["Lines -> Tubes"]) 
    l2t_menu.add_item(Lines2Tubes::L2T["Set profile"]) { Lines2Tubes.setFaceProfile() }
    l2t_menu.add_item(Lines2Tubes::L2T["Create tubes"]) { Lines2Tubes.cmdlines2tubes() }
    
    # Toolbar
	l2t_tb = UI::Toolbar.new(Lines2Tubes::L2T["Lines to tubes"])
	
	cmd1 = UI::Command.new(Lines2Tubes::L2T["Set profile"]) { Lines2Tubes.setFaceProfile() } 
	cmd1.small_icon = "Resources/L2T1.png"
	cmd1.large_icon = "Resources/L2T1.png"
	cmd1.tooltip = Lines2Tubes::L2T["Set profile"]
	cmd1.status_bar_text=Lines2Tubes::L2T["Use selected face as extrusion profile"]
	l2t_tb.add_item(cmd1)
	
	cmd0 = UI::Command.new(Lines2Tubes::L2T["Create tubes"]) { Lines2Tubes.cmdlines2tubes } 
	cmd0.small_icon = "Resources/L2T2.png"
	cmd0.large_icon = "Resources/L2T2.png"
	cmd0.tooltip = Lines2Tubes::L2T["Lines to tubes"]
	cmd0.status_bar_text=Lines2Tubes::L2T["Convert selected arcs, circles, curves, edges to tubes"]
	l2t_tb.add_item(cmd0)
	
	status=l2t_tb.get_last_state
	l2t_tb.show() if status<1
end
file_loaded("lines2tubes.rb")


end # module Lines2Tubes
